"""Logging fixtures and functions."""
import contextlib
import logging
import logging.handlers
import socket

import pytest


@pytest.fixture(scope='session')
def services_log(worker_id):
    """A services_logger with the worker id."""
    handler = None
    for kwargs in (dict(socktype=socket.SOCK_RAW), dict(socktype=socket.SOCK_STREAM), dict()):
        try:
            handler = logging.handlers.SysLogHandler(
                facility=logging.handlers.SysLogHandler.LOG_LOCAL7, address='/dev/log', **kwargs)
            break
        except (IOError, TypeError):
            pass
    logger = logging.getLogger('[{worker_id}] {name}'.format(name=__name__, worker_id=worker_id))
    logger.setLevel(logging.DEBUG)
    if handler and workaround_issue_20(handler):
        logger.propagate = 0
        logger.addHandler(handler)
    return logger


def workaround_issue_20(handler):
    """
    Workaround for
    https://github.com/pytest-dev/pytest-services/issues/20,
    disabling installation of a broken handler.
    """
    return hasattr(handler, 'socket')


@contextlib.contextmanager
def dont_capture(request):
    """Suspend capturing of stdout by pytest."""
    capman = request.config.pluginmanager.getplugin("capturemanager")
    capman.suspendcapture()
    try:
        yield
    finally:
        capman.resumecapture()


def remove_handlers():
    """Remove root services_logging handlers."""
    handlers = []
    for handler in logging.root.handlers:
        if not isinstance(handler, logging.StreamHandler):
            handlers.append(handler)
    logging.root.handlers = handlers



"""Fixtures for mysql."""
import os
import shutil
from textwrap import dedent

import pytest

from .process import (
    CalledProcessWithOutputError,
    check_output,
)


@pytest.fixture(scope='session')
def mysql_defaults_file(
        run_services, tmp_path_factory, memory_temp_dir, request):
    """MySQL defaults file."""
    if run_services:
        cfg = tmp_path_factory.mktemp("pytest-services")
        defaults_path = str(cfg / 'defaults.cnf')

        with open(defaults_path, 'w+') as fd:
            user = os.environ["USER"]
            fd.write(
                dedent(
                    f"""
                    [mysqld]
                    user = {user}
                    tmpdir = {memory_temp_dir}
                    default-time-zone = SYSTEM
                    """
                )
            )
        return defaults_path
    return None


@pytest.fixture(scope='session')
def mysql_base_dir():
    my_print_defaults = shutil.which('my_print_defaults')
    assert my_print_defaults, 'You have to install my_print_defaults script.'

    return os.path.dirname(os.path.dirname(os.path.realpath(my_print_defaults)))


@pytest.fixture(scope='session')
def mysql_system_database(
        run_services,
        mysql_data_dir,
        mysql_base_dir,
        mysql_defaults_file,
        memory_temp_dir,
        lock_dir,
        services_log,
):
    """Install database to given path."""
    if run_services:
        pytest.skip(reason="#50 needs investigation")
        mysqld = shutil.which('mysqld')
        assert mysqld, 'You have to install mysqld script.'

        try:
            services_log.debug('Starting mysqld.')
            check_output([
                mysqld,
                '--defaults-file={0}'.format(mysql_defaults_file),
                '--initialize-insecure',
                '--datadir={0}'.format(mysql_data_dir),
                '--basedir={0}'.format(mysql_base_dir),
                '--user={0}'.format(os.environ['USER'])
            ])
        except CalledProcessWithOutputError as e:
            services_log.error(
                '{e.cmd} failed with output:\n{e.output}\nand erorr:\n{e.err}. '
                'Please ensure you disabled apparmor for /run/shm/** or for whole mysql'.format(e=e))
            raise
        finally:
            services_log.debug('mysqld was executed.')


@pytest.fixture(scope='session')
def mysql_data_dir(
        request, memory_base_dir, memory_temp_dir, lock_dir, session_id, services_log, run_services):
    """The root directory for the mysql instance.

    `mysql_install_db` is run in that directory.

    """
    if run_services:
        path = os.path.join(memory_base_dir, 'mysql')
        services_log.debug('Making mysql base dir in {path}'.format(path=path))

        def finalizer():
            shutil.rmtree(path, ignore_errors=True)

        finalizer()
        request.addfinalizer(finalizer)
        os.mkdir(path)
        return path


@pytest.fixture(scope='session')
def mysql_socket(run_dir):
    """The mysqld socket location."""
    return os.path.join(run_dir, 'mysql.sock')


@pytest.fixture(scope='session')
def mysql_pid(run_dir):
    """The pid file of the mysqld."""
    return os.path.join(run_dir, 'mysql.pid')


@pytest.fixture(scope='session')
def mysql_connection(run_services, mysql_socket):
    """The connection string to the local mysql instance."""
    if run_services:
        return 'mysql://root@localhost/?unix_socket={0}&charset=utf8'.format(mysql_socket)


@pytest.fixture(scope='session')
def mysql_watcher(
        request, run_services, watcher_getter, mysql_system_database, mysql_pid, mysql_socket, mysql_data_dir,
        mysql_defaults_file):
    """The mysqld process watcher."""
    if run_services:
        return watcher_getter(
            'mysqld',
            [
                '--defaults-file={0}'.format(mysql_defaults_file),
                '--datadir={mysql_data_dir}'.format(mysql_data_dir=mysql_data_dir),
                '--pid-file={mysql_pid}'.format(mysql_pid=mysql_pid),
                '--socket={mysql_socket}'.format(mysql_socket=mysql_socket),
                '--skip-networking',
            ],
            checker=lambda: os.path.exists(mysql_socket),
            request=request,
        )


@pytest.fixture(scope='session')
def mysql_database_name():
    """Name of test database to be created."""
    return 'pytest_services_test'


@pytest.fixture(scope='session')
def mysql_database_getter(run_services, mysql_watcher, mysql_socket):
    """Prepare new test database creation function."""
    if run_services:
        def getter(database_name):
            check_output(
                [
                    'mysql',
                    '--user=root',
                    '--socket={0}'.format(mysql_socket),
                    '--execute=create database {0};'.format(database_name),
                ],
            )
            check_output(
                'mysql_tzinfo_to_sql /usr/share/zoneinfo | mysql --user=root --socket={0} mysql'.format(mysql_socket),
                shell=True
            )
        return getter


@pytest.fixture(scope='session')
def mysql_database(run_services, mysql_database_getter, mysql_database_name):
    """Prepare new test database creation function."""
    if run_services:
        return mysql_database_getter(mysql_database_name)


@pytest.fixture(scope='session')
def mysql(request, run_services, mysql_watcher, mysql_database):
    """The mysql instance which is ready to be used by the tests."""
    if run_services:
        return mysql_watcher



"""Subprocess related functions."""
try:
    import subprocess32 as subprocess
except ImportError:
    import subprocess


def check_output(*popenargs, **kwargs):
    """Run command with arguments and return its output (both stdout and stderr) as a byte string.

    If the exit code was non-zero it raises a CalledProcessWithOutputError.
    """
    if 'stdout' in kwargs:
        raise ValueError('stdout argument not allowed, it will be overridden.')

    if 'stderr' in kwargs:
        raise ValueError('stderr argument not allowed, it will be overridden.')

    process = subprocess.Popen(
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        *popenargs, **kwargs)
    output, err = process.communicate()
    retcode = process.poll()
    if retcode:
        cmd = kwargs.get("args")
        if cmd is None:
            cmd = popenargs[0]
        raise CalledProcessWithOutputError(retcode, cmd, output, err)

    return output, err


class CalledProcessWithOutputError(subprocess.CalledProcessError):

    """An exception with the steout and stderr of a failed subprocess32."""

    def __init__(self, returncode, cmd, output, err):
        """Assign output and error."""
        super(CalledProcessWithOutputError, self).__init__(returncode, cmd)

        self.output = output
        self.err = err

    def __str__(self):
        """str representation."""
        return super(CalledProcessWithOutputError, self).__str__() + ' with output: {0} and error: {1}'.format(
            self.output, self.err)



"""Services plugin for pytest.

Provides an easy way of running service processes for your tests.
"""

from .folders import *  # NOQA
from .log import *  # NOQA
from .locks import *  # NOQA
from .xvfb import *  # NOQA
from .memcached import *  # NOQA
from .mysql import *  # NOQA
from .service import *  # NOQA


def pytest_addoption(parser):
    """Add options for services plugin."""
    group = parser.getgroup("services", "service processes for tests")
    group._addoption(
        '--run-services',
        action="store_true", dest="run_services",
        default=False,
        help="Run services automatically by pytest")
    group._addoption(
        '--xvfb-display',
        action="store", dest="display",
        default=None,
        help="X display to use")



"""Fixtures for the GUI environment."""
import os
import socket
import re
try:
    import subprocess32 as subprocess
except ImportError:  # pragma: no cover
    import subprocess

import pytest

from .locks import (
    file_lock,
)


def xvfb_supports_listen():
    """Determine whether the '-listen' option is supported by Xvfb."""
    p = subprocess.Popen(
        ['Xvfb', '-listen', 'TCP', '-__sentinel_parameter__'],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    p.wait()
    _, stderr = p.communicate()

    match = re.search(
        br'^Unrecognized option: (?P<option>.*)$',
        stderr,
        flags=re.MULTILINE,
    ).groupdict()
    unrecognized_option = match['option']
    return unrecognized_option != b'-listen'


@pytest.fixture(scope='session')
def xvfb_display(request, run_services, lock_dir, services_log, display_getter):
    """The DISPLAY environment variable used in this test run.

    In case it is not a local run, a random value will be picked up and set,
    otherwise it will be taken from the environment.

    """
    if run_services:
        if request.config.option.display:
            display = request.config.option.display
        else:
            display = display_getter()
        os.environ['DISPLAY'] = ':{0}'.format(display) if ':' not in str(display) else display

        return display


@pytest.fixture(scope='session')
def xvfb_resolution():
    """xvfb display resolution."""
    return (1366, 768, 8)


@pytest.fixture(scope='session')
def xvfb(request, run_services, xvfb_display, lock_dir, xvfb_resolution, watcher_getter):
    """The Xvfb process."""
    if request.config.option.display or not run_services:
        # display is passed, no action required
        return

    if xvfb_supports_listen():
        listen_args = ['-listen', 'TCP']
    else:
        listen_args = []

    with file_lock(os.path.join(lock_dir, 'xvfb_{0}.lock'.format(xvfb_display)),
                   ):

        def checker():
            try:
                socket.create_connection(('127.0.0.1', 6000 + xvfb_display))
                return True
            except socket.error:
                pass

        return watcher_getter(
            'Xvfb', [
                ':{display}'.format(display=xvfb_display),
                '-screen',
                '0',
                'x'.join(str(value) for value in xvfb_resolution),
                '-ac',
                '-nolock',
                '+extension', 'RANDR'
            ] + listen_args,
            checker=checker,
            request=request,
        )



"""Service fixtures."""
import time
import re
import warnings
try:
    import subprocess32 as subprocess
except ImportError:  # pragma: no cover
    import subprocess
import uuid  # pylint: disable=C0411

from shutil import which
import pytest

WRONG_FILE_NAME_CHARS_RE = re.compile(r'[^\w_-]')


@pytest.fixture(scope='session')
def run_services(request, worker_id):
    """Indicate whether the services should run or not."""
    return worker_id != 'local' or request.config.option.run_services


@pytest.fixture(scope='session')
def worker_id(request):
    """
    The id of the worker if tests are run using xdist.

    It is set to `'local'` if tests are not run using xdist.

    The id is unique for a test run. An id may clash if there are two workers
    that belong to different test sessions.

    """
    return WRONG_FILE_NAME_CHARS_RE.sub('_', getattr(request.config, 'workerinput', {}).get('workerid', 'local'))


@pytest.fixture(scope='session')
def slave_id(request, worker_id):
    msg = "The `slave_id` fixture is deprecated; use `worker_id` instead."
    warnings.warn(msg, DeprecationWarning)
    return worker_id


@pytest.fixture(scope='session')
def session_id(request, worker_id, run_services):
    """The test session id.

    It is supposed to be globally unique.

    """
    # UUID should be enough, other fields are added for the debugging purposes.
    session_id = '{random}-{worker_id}'.format(
        random=uuid.uuid4().hex,
        worker_id=worker_id,
    )

    return session_id


@pytest.fixture(scope='session')
def watcher_getter(request, services_log):
    """Popen object of given executable name and it's arguments.

    Wait for the process to start.
    Add finalizer to properly stop it.
    """
    orig_request = request

    def watcher_getter_function(name, arguments=None, kwargs=None, timeout=20, checker=None, request=None):
        if request is None:
            warnings.warn('Omitting the `request` parameter will result in an unstable order of finalizers.')
            request = orig_request
        executable = which(name)
        assert executable, 'You have to install {0} executable.'.format(name)

        cmd = [name] + (arguments or [])

        services_log.debug('Starting {0}: {1}'.format(name, arguments))

        watcher = subprocess.Popen(
            cmd, **(kwargs or {}))

        def finalize():
            try:
                watcher.terminate()
            except OSError:
                pass
            if watcher.returncode is None:
                try:
                    watcher.communicate(timeout=timeout / 2)
                except subprocess.TimeoutExpired:
                    watcher.kill()
                    watcher.communicate(timeout=timeout / 2)
        request.addfinalizer(finalize)

        # Wait for the service to start.
        times = 0
        while not checker():
            if watcher.returncode is not None:
                raise Exception("Error running {0}".format(name))

            if times > timeout:
                raise Exception('The {0} service checked did not succeed!'.format(name))

            time.sleep(1)
            times += 1

        return watcher

    return watcher_getter_function
